# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['darglint',
 'darglint.analysis',
 'darglint.docstring',
 'darglint.parse',
 'darglint.parse.grammars']

package_data = \
{'': ['*']}

entry_points = \
{'console_scripts': ['darglint = darglint.driver:main'],
 'flake8.extension': ['DAR = darglint.flake8_entry:DarglintChecker']}

setup_kwargs = {
    'name': 'darglint',
    'version': '1.8.0',
    'description': 'A utility for ensuring Google-style docstrings stay up to date with the source code.',
    'long_description': '[![Build Status](https://travis-ci.com/terrencepreilly/darglint.svg?branch=develop)](https://travis-ci.com/terrencepreilly/darglint)\n\n# Darglint\n\nA functional docstring linter which checks whether a docstring\'s\ndescription matches the actual function/method implementation.\n*Darglint* expects docstrings to be formatted using the\n[Google Python Style Guide](https://google.github.io/styleguide/pyguide.html),\nor [Sphinx Style Guide](https://pythonhosted.org/an_example_pypi_project/sphinx.html#function-definitions),\nor [Numpy Style Guide](https://numpydoc.readthedocs.io/en/latest/format.html).\n\nFeel free to submit an issue/pull request if you spot a problem or\nwould like a feature in *darglint*.\n\n**Table of Contents**:\n\n- [Installation](#installation)\n- [Configuration](#configuration)\n- [Usage](#usage)\n- [Scope](#scope)\n- [Sphinx](#sphinx)\n- [Numpy](#numpy)\n- [Integrations](#integrations)\n- [Flake8](#flake8)\n- [Roadmap](#roadmap)\n- [Contribution](#development-and-contributions)\n\n\n## Installation\n\nTo install *darglint*, use pip.\n\n```bash\npip install darglint\n```\n\nOr, clone the repository, `cd` to the directory, and\n\n```bash\npip install .\n```\n\n## Configuration\n\n*darglint* can be configured using a configuration file.  The configuration\nfile must be named either *.darglint*, *setup.cfg*, or *tox.ini*.  It must\nalso have a section starting with the section header, `[darglint]`.\nFinally, the configuration file must be located either in the directory\n*darglint* is called from, or from a parent directory of that working\ndirectory.\n\nCurrently, the configuration file allows us to ignore errors, to specify\nmessage templates, to specify the strictness of checks and to ignore common\nexceptions.\n\n### Error Configuration\n\nIf we would like to ignore `ExcessRaiseError`s (because we know that\nan underlying function will raise an exception), then we would add its\nerror code to a file named *.darglint*:\n\n```ini\n[darglint]\nignore=DAR402\n```\n\nWe can ignore multiple errors by using a comma-separated list:\n\n```ini\n[darglint]\nignore=DAR402,DAR103\n```\n\nInstead of specifying error codes to ignore in general one can also specify a\nregex to exclude certain function names from tests. For example, the following \nconfiguration would disable linting on all private methods.\n```ini\n[darglint]\nignore_regex=^_(.*)\n```\n\n### Message Template Configuration\n\nIf we would like to specify a message template, we may do so as\nfollows:\n\n```ini\n[darglint]\nmessage_template={msg_id}@{path}:{line}\n```\n\nWhich will produce a message such as `DAR102@driver.py:72`.\n\nFinally, we can specify the docstring style type using `docstring_style`\n("google" by default):\n\n```ini\n[darglint]\ndocstring_style=sphinx\n```\n\n### Strictness Configuration\n\nStrictness determines how lax darglint will be when checking docstrings.\nThere are three levels of strictness available:\n\n- short: One-line descriptions are acceptable; anything\nmore and the docstring will be fully checked.\n\n- long: One-line descriptions and descriptions without\narguments/returns/yields/etc. sections will be allowed.  Anything more,\nand the docstring will be fully checked.\n\n- full: (Default) Docstrings will be fully checked.\n\nFor example, if we have the following function:\n\n```python\ndef double(x):\n    # <docstring>\n    return x * 2\n```\n\nThen the following table describes which errors will be raised for\neach of the docstrings (rows) when checked against each of the\nconfigurations (columns):\n\n```\n┌──────────────────────────────┬──────────────────┬────────────────┬──────────────────┐\n│ Docstring                    │  short           │  long          │  full            │\n├──────────────────────────────┼──────────────────┼────────────────┼──────────────────┤\n│ """Doubles the argument."""  │ None             │ None           │ Missing argument │\n│                              │                  │                │ Missing return   │\n│                              │                  │                │                  │\n│                              │                  │                │                  │\n├──────────────────────────────┼──────────────────┼────────────────┼──────────────────┤\n│ """Doubles the argument.     │ Missing argument │ None           │ Missing argument │\n│                              │ Missing return   │                │ Missing return   │\n│ Not very pythonic.           │                  │                │                  │\n│                              │                  │                │                  │\n│ """                          │                  │                │                  │\n│                              │                  │                │                  │\n├──────────────────────────────┼──────────────────┼────────────────┼──────────────────┤\n│ """Doubles the argument.     │ Missing return   │ Missing return │ Missing return   │\n│                              │                  │                │                  │\n│ Args:                        │                  │                │                  │\n│     x: The number to double. │                  │                │                  │\n│                              │                  │                │                  │\n│ """                          │                  │                │                  │\n└──────────────────────────────┴──────────────────┴────────────────┴──────────────────┘\n```\n\nIn short, if you want to be able to have single-line docstrings, and check\nall other docstrings against their described parameters, you would specify\n\n```ini\n[darglint]\nstrictness=short\n```\n\nIn your configuration file.\n\n### Ignoring common exceptions\n\nWe can specify a list of exceptions that don\'t need to be documented in the\nraises section of a docstring. For example,\n\n```ini\n[darglint]\nignore_raise=ValueError,MyCustomError\n```\n\n### Logging\n\nWhen *darglint* fails unexpectedly, you can try to gather more\ninformation when submitting a bug by running with logging.\nFor example,\n\n```bash\ndarglint --log-level=INFO unexpected_failures.py\n```\n\n*Darglint* accepts the levels, `DEBUG`, `INFO`, `WARNING`, `ERROR`, and\n`CRITICAL`.\n\n\n## Usage\n\n\n### Command Line use\n\nGiven a python source file, `serializers.py`, you would check the docstrings\nas follows:\n\n```bash\ndarglint serializers.py\n```\n\nYou can give an optional verbosity setting to *darglint*.  For example,\n\n```bash\ndarglint -v 2 *.py\n```\n\nWould give a description of the error along with information as to this\nspecific instance.  The default verbosity is 1, which gives the filename,\nfunction name, line number, error code, and some general hints.\n\nTo use an arbitrary error format, you can pass a message template, which\nis a python format string.  For example, if we pass the message\ntemplate\n\n```bash\ndarglint -m "{path}:{line} -> {msg_id}" darglint/driver.py\n```\n\nThen we would get back error messages like\n\n```bash\ndarglint/driver.py :61 -> DAR101\n```\n\nThe following attributes can be passed to the format string:\n- *line*: The line number,\n- *msg*: The error message,\n- *msg_id*: The error code,\n- *obj*: The function/method name,\n- *path*: The relative file path.\n\nThe message template can also be specified in the configuration file\nas the value `message_template`.\n\n*darglint* is particularly useful when combined with the utility, `find`.\nThis allows us to check all of the files in our project at once.  For example,\nwhen eating my own dogfood (as I tend to do), I invoke *darglint* as follows:\n\n```bash\nfind . -name "*.py" | xargs darglint\n```\n\nWhere I\'m searching all files ending in ".py" recursively from the\ncurrent directory, and calling *darglint* on each one in turn.\n\n### Ignoring Errors in a Docstring\n\nYou can ignore specific errors in a particular docstring.  The syntax\nis much like that of *pycodestyle*, etc.  It generally takes the from\nof:\n\n```python\n# noqa: <error> <argument>\n```\n\nWhere `<error>` is the particular error to ignore (`DAR402`, or `DAR201`\nfor example), and `<argument>` is what (if anything) the ignore\nstatement refers to (if nothing, then it is not specified).\n\nLet us say that we want to ignore a missing return statement\nin the following docstring:\n\n```python\ndef we_dont_want_a_returns_section():\n  """Return the value, 3.\n\n  # noqa: DAR201\n\n  """\n  return 3\n```\n\nWe put the `noqa` anywhere in the top level of the docstring.\nHowever, this won\'t work if we are missing something more specific,\nlike a parameter.  We may not want to ignore all missing parameters,\neither, just one particular one.  For example, we may be writing a\nfunction that takes a class instance as self. (Say, in a bound *celery*\ntask.) Then we would do something like:\n\n```python\ndef a_bound_function(self, arg1):\n  """Do something interesting.\n\n  Args:\n    arg1: The first argument.\n\n  # noqa: DAR101 arg1\n\n  """\n  arg1.execute(self)\n```\n\nSo, the argument comes to the right of the error.\n\nWe may also want to mark excess documentation as being okay.  For example,\nwe may not want to explicitly catch and raise a `ZeroDivisionError`.  We\ncould do the following:\n\n```python\ndef always_raises_exception(x):\n    """Raise a zero division error or type error.o\n\n    Args:\n      x: The argument which could be a number or could not be.\n\n    Raises:\n      ZeroDivisionError: If x is a number.  # noqa: DAR402\n      TypeError: If x is not a number.  # noqa: DAR402\n\n    """\n    x / 0\n```\n\nSo, in this case, the argument for `noqa` is really all the way to\nthe left.  (Or whatever description we are parsing.)  We could also\nhave put it on its own line, as `# noqa: DAR402 ZeroDivisionError`.\n\n### Type Annotations\n\nDarglint parses type annotations in docstrings, and can, optionally,\ncompare the documented type to the actual type annotation.  This can\nbe useful when migrating a codebase to use type annotations.\n\nIn order to make these comparisons, Darglint only accepts types\naccepted by Python (see [PEP 484](https://www.python.org/dev/peps/pep-0484/).)\nThat is, it does not accept parentheses in type signatures. (If\nparentheses are used in the type signature, Darglint will mark that\nargument as missing.  See Issue #90.)\n\n\n### Error Codes\n\n- *DAR001*: The docstring was not parsed correctly due to a syntax error.\n- *DAR002*: An argument/exception lacks a description\n- *DAR003*: A line is under-indented or over-indented.\n- *DAR004*: The docstring contains an extra newline where it shouldn\'t.\n- *DAR005*: The item contains a type section (parentheses), but no type.\n- *DAR101*: The docstring is missing a parameter in the definition.\n- *DAR102*: The docstring contains a parameter not in function.\n- *DAR103*: The docstring parameter type doesn\'t match function.\n- *DAR104*: (disabled) The docstring parameter has no type specified \n- *DAR105*: The docstring parameter type is malformed.\n- *DAR201*: The docstring is missing a return from definition.\n- *DAR202*: The docstring has a return not in definition.\n- *DAR203*: The docstring parameter type doesn\'t match function.\n- *DAR301*: The docstring is missing a yield present in definition.\n- *DAR302*: The docstring has a yield not in definition.\n- *DAR401*: The docstring is missing an exception raised.\n- *DAR402*: The docstring describes an exception not explicitly raised.\n- *DAR501*: The docstring describes a variable which is not defined.\n\nThe number in the hundreds narrows the error by location in the docstring:\n\n- 000: Syntax, formatting, and style\n- 100: Args section\n- 200: Returns section\n- 300: Yields section\n- 400: Raises section\n- 500: Variables section\n\nYou can enable disabled-by-default exceptions in the configuration file\nusing the `enable` option.  It accepts a comma-separated list of error\ncodes.\n\n```ini\n[darglint]\nenable=DAR104\n```\n\n## Scope\n\nDarglint\'s primary focus is to identify incorrect and missing documentationd\nof a function\'s signature. Checking style is a stretch goal, and is supported\non a best-effort basis.  Darglint does not check stylistic preferences expressed\nby tools in the Python Code Quality Authority (through tools such as `pydocstyle`).\nSo when using Darglint, it may be a good idea to also use `pydocstyle`, if you\nwant to enforce style.  (For example, `pydocstyle` requires the short summary\nto be separated from other sections by a line break.  Darglint makes no such check.)\n\n## Sphinx\n\nDarglint can handle sphinx-style docstrings, but imposes some restrictions\non top of the Sphinx style. For example, all fields (such as `:returns:`)\nmust be the last items in the docstring.  They must be together, and all\nindents should be four spaces.  These restrictions may be loosened at a\nlater date.\n\nTo analyze Sphinx-style docstrings, pass the style flag to the command:\n\n```bash\ndarglint -s sphinx example.py\ndarglint --docstring-style sphinx example.py\n```\n\nAlternatively, you can specify the style in the configuration file using\nthe setting, "docstring\\_style":\n\n```ini\n[darglint]\ndocstring_style=sphinx\n```\n\n## Numpy\n\nDarglint now has an initial implementation for Numpy-style docstrings.\nSimilarly to Sphinx-style docstrings, you can pass a style flag to the\ncommand:\n\n```bash\ndarglint -s numpy example.py\ndarglint --docstring-style numpy example.py\n```\n\nOr set it in a configuration file:\n\n```ini\n[darglint]\ndocstring_style=numpy\n```\n\nThe numpy parser and error reporter are not yet fully stabilized.\nAdd issues or suggestions to the tracking bug, Issue #69.\n\n## Integrations\n\n### Flake8\n\nDarglint can be used in conjunction with Flake8 as a plugin.  The only\nsetup necessary is to install Flake8 and Darglint in the same environment.\nDarglint will pull its configuration from Flake8. So, if you would like to\nlint Sphinx-style comments, then you should have `docstring_style=sphinx` in a\nFlake8 configuration file in the project directory.  The settings would\nbe entered under the flake8 configuration, not a separate configuration\nfor Darglint.  E.g.:\n\n```ini\n[flake8]\nstrictness=short\ndocstring_style=sphinx\n```\n\nTo see which options are exposed through Flake8, you can check the Flake8\ntool:\n\n```bash\nflake8 --help | grep --before-context=2 Darglint\n```\n\n### SublimeLinter\n\nA plugin for SublimeLinter can be found [here](https://github.com/raddessi/SublimeLinter-contrib-darglint)\n\n### Pre-commit\n\nDownload [pre-commit](https://pre-commit.com/) and\n[install](https://pre-commit.com/#install) it. Once it is installed, add this\nto `.pre-commit-config.yaml` in your repository:\n\n```yaml\nrepos:\n-   repo: https://github.com/terrencepreilly/darglint\n    rev: master\n    hooks:\n    - id: darglint\n```\n\nThen run `pre-commit install` and you\'re ready to go. Before commiting,\n`darglint` will be run on the staged files. If it finds any errors, the user\nis notified and the commit is aborted. Store necessary configuration (such as\nerror formatting) in `.darglint`, `setup.cfg` or `tox.ini`.\n\n\n## Roadmap\n\nBelow are some of the current features or efforts.  Where a milestone or\nissue is associated with the idea, it will be mentioned.  Some of these\nideas are moonshots and may not get implemented.  They are ordered\nroughly according to current priority/feasibility.\n\n- [ ] Expose command-line options through sphinx.\n- [ ] Robust logging for errors caused/encountered by *darglint*.\n- [ ] Check class docstrings (See Issue #25).\n- [ ] Autoformatting docstrings.  (See Milestone #3).\n- [ ] Optional aggressive style checking through command line flag.\n- [ ] ALE support.\n- [ ] Syntastic support. (Syntastic is not accepting new checkers until\ntheir next API stabilizes, so this may take some time.)\n\n\n## Development and Contributions\n\n### Development Setup\n\nInstall `darglint`. First, clone the repository:\n\n```bash\ngit clone https://github.com/terrencepreilly/darglint.git\n```\n\n`cd` into the directory, create a virtual environment (optional), then setup:\n\n```bash\ncd darglint/\nvirtualenv -p python3.6 .env\nsource .env/bin/activate\npip install -e .\n```\n\nYou can install dependencies using\n\n```bash\npip install poetry\npoetry install\n```\n\nYou can run the tests using\n\n```bash\npython setup.py test\n```\n\nOr, install `pytest` manually, `cd` to the project\'s root directory,\nand run\n\n```bash\npytest\n```\n\nThis project tries to conform by the styles imposed by `pycodestyle`\nand `pydocstyle`, as well as by `darglint` itself.\n\n\nA dockerfile exists for testing with Python3.4.  Although it\'s not\nofficially supported (only 3.6+), it\'s nice to try to make minor\nversion numbers support it.  You would build the dockerfile and\ntest using something like\n\n```bash\npushd docker-build\ndocker build -t darglint-34 -f Dockerfile.test34 .\npopd\ndocker run -it --rm -v $(pwd):/code darglint-34 pytest\n```\n\n### Contribution\n\nIf you would like to tackle an issue or feature, email me or comment on the\nissue to make sure it isn\'t already being worked on.  Contributions will\nbe accepted through pull requests.  New features should include unit tests,\nand, of course, properly formatted documentation.\n\nAlso, check out the wiki prior to updating the grammar.  It includes a\ndescription of darglint\'s parsing pipline.\n',
    'author': 'terrencepreilly',
    'author_email': 'terrencepreilly@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import ast
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)
from ..config import (
    get_logger,
)
from ..custom_assert import (
    Assert,
)


class ReturnVisitor(ast.NodeVisitor):
    """A visitor which checks for *returns* nodes."""

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # TODO: https://github.com/python/mypy/issues/4001
        super(ReturnVisitor, self).__init__(*args, **kwargs)  # type: ignore

        # A list of the return nodes encountered.
        self.returns = list()  # type: List[Optional[ast.Return]]
        self.return_types = list()  # type: List[Optional[ast.AST]]

    def visit_Return(self, node):
        # type: (ast.Return) -> ast.AST
        self.returns.append(node)
        return self.generic_visit(node)



from collections import defaultdict
from typing import (  # noqa
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

from .base import BaseDocstring
from .sections import Sections
from .style import DocstringStyle
from ..custom_assert import Assert
from ..node import (
    CykNode,
)
from ..parse.identifiers import (
    Identifier,
    NoqaIdentifier,
)
from ..parse.sphinx import (
    parse,
)
from ..lex import (
    lex,
    condense,
)
from ..strictness import Strictness
from ..errors import (
    DarglintError,
)


class Docstring(BaseDocstring):
    """The docstring class interprets the AST of a docstring."""

    def __init__(self, root, style=DocstringStyle.SPHINX):
        # type: (Union[CykNode, str], DocstringStyle) -> None  # noqa: E501
        """Create a new docstring from the AST.

        Args:
            root: The root of the AST, or the docstring
                (as a string.)  If it is a string, the
                string will be parsed.
            style: The docstring style.  Discarded, since this
                docstring always represents the Sphinx style.

        """
        if isinstance(root, CykNode):
            self.root = root
        else:
            self.root = parse(condense(lex(root)))
        self._lookup = self._discover()

    def _discover(self):
        # type: () -> Dict[str, List[CykNode]]
        """Walk the tree, finding all non-terminal nodes.

        Returns:
            A lookup table for compound Nodes by their NodeType.

        """
        lookup = defaultdict(
            lambda: list()
        )  # type: Dict[str, List[CykNode]]
        for node in self.root.in_order_traverse():
            lookup[node.symbol].append(node)
            for annotation in node.annotations:
                if issubclass(annotation, Identifier):
                    lookup[annotation.key].append(node)
        return lookup

    def get_section(self, section):
        # type: (Sections) -> Optional[str]
        nodes = []  # type: Optional[List[CykNode]]

        if section == Sections.SHORT_DESCRIPTION:
            nodes = self._lookup.get('short-description', None)
        elif section == Sections.LONG_DESCRIPTION:
            nodes = self._lookup.get('long-description', None)
        elif section == Sections.ARGUMENTS_SECTION:
            nodes = self._lookup.get('arguments-section', None)
        elif section == Sections.RAISES_SECTION:
            nodes = self._lookup.get('raises-section', None)
        elif section == Sections.YIELDS_SECTION:
            nodes = self._lookup.get('yields-section', None)
        elif section == Sections.RETURNS_SECTION:
            nodes = self._lookup.get('returns-section', None)
        elif section == Sections.VARIABLES_SECTION:
            nodes = self._lookup.get('variables-section', None)
        elif section == Sections.NOQAS:
            nodes = self._lookup.get('noqa', None)
        else:
            raise Exception(
                'Unsupported section type, {}'.format(section)
            )

        if not nodes:
            return None

        return_value = ''
        for node in nodes:
            return_value += '\n\n' + node.reconstruct_string()

        return return_value.strip() or None

    def _get_argument_type_lookup(self):
        # type: () -> Dict[str, Optional[str]]
        ret = dict()  # type: Dict[str, Optional[str]]
        for section in self._lookup['arguments-section']:

            Assert(section.lchild, 'Section unexpected had no left child.')
            if section.lchild is None:
                continue

            argument = section.lchild.first_instance('word')
            if argument and argument.value:
                ret[argument.value.value] = None
        # TODO: Change to use @ArgumentTypeIdentifier
        for argtype in self._lookup['argument-type-section']:
            if argtype.lchild:
                word = argtype.lchild.first_instance('word')
                if word:
                    argument_type = None
                    if argtype.rchild:
                        argument_type = (
                            argtype.rchild.reconstruct_string().strip()
                        )
                    Assert(word.value, 'Word unexpectedly had no value')
                    if not word.value:
                        continue
                    ret[word.value.value] = argument_type or None
        return ret

    def _get_raises_type(self):
        # type: () -> List[Optional[str]]
        ret = list()  # type: List[Optional[str]]
        for section in self._lookup['raises-section']:
            Assert(section.lchild, 'Section unexpected had no left child.')
            if section.lchild is None:
                continue
            exception = section.lchild.first_instance('word')
            if exception and exception.value:
                ret.append(exception.value.value)
            else:
                ret.append(None)
        return ret

    def _get_variable_type_lookup(self):
        # type: () -> Dict[str, Optional[str]]
        ret = defaultdict()  # type: Dict[str, Optional[str]]
        for section in self._lookup['variables-section']:
            Assert(section.lchild, 'Section unexpected had no left child.')
            if section.lchild is None:
                continue
            variable = section.lchild.first_instance('word')
            if variable and variable.value:
                ret[variable.value.value] = None
        for section in self._lookup['variable-type-section']:
            Assert(section.lchild, 'Section unexpected had no left child.')
            if section.lchild is None:
                continue
            variable = section.lchild.first_instance('word')
            if variable and variable.value:
                Assert(
                    section.rchild,
                    'Section unexpected had no right child.'
                )
                if section.rchild is None:
                    continue
                vartype = section.rchild.reconstruct_string().strip()
                ret[variable.value.value] = vartype
        return ret

    def _get_return_type(self):
        # type: () -> Optional[str]
        if 'return-type-section' not in self._lookup:
            return None
        return_type_section = self._lookup['return-type-section'][0]

        Assert(
            return_type_section.rchild,
            'Return type unexpectedly had no right child.'
        )
        if not return_type_section.rchild:
            return None
        return return_type_section.rchild.reconstruct_string().strip()

    def _get_yield_type(self):
        # type: () -> Optional[str]
        if 'yield-type-section' not in self._lookup:
            return None
        yield_type_section = self._lookup['yield-type-section'][0]

        Assert(
            yield_type_section.rchild,
            'Yield type unexpectedly had not right child.'
        )
        if not yield_type_section.rchild:
            return None
        return yield_type_section.rchild.reconstruct_string().strip()

    def _sorted_values(self, lookup):
        # type: (Dict[str, Optional[str]]) -> List[Optional[str]]
        return [lookup[key] for key in sorted(lookup.keys())]

    def _sorted_keys(self, lookup):
        return sorted(lookup.keys())

    def get_types(self, section):
        # type: (Sections) -> Optional[Union[str, List[Optional[str]]]]
        if section == Sections.ARGUMENTS_SECTION:
            if 'arguments-section' not in self._lookup:
                return None
            return (
                self._sorted_values(self._get_argument_type_lookup())
                or None
            )
        elif section == Sections.VARIABLES_SECTION:
            if 'variables-section' not in self._lookup:
                return None
            return (
                self._sorted_values(self._get_variable_type_lookup())
                or None
            )
        elif section == Sections.RETURNS_SECTION:
            return self._get_return_type() or None
        elif section == Sections.YIELDS_SECTION:
            return self._get_yield_type() or None
        else:
            raise Exception(
                'Section type {} does not have types, '.format(
                    section.name
                ) + 'or is not yet supported'
            )
        return None

    def get_items(self, section):
        # type: (Sections) -> Optional[List[str]]
        if section == Sections.ARGUMENTS_SECTION:
            return self._sorted_keys(self._get_argument_type_lookup()) or None
        elif section == Sections.RAISES_SECTION:
            return sorted(
                [x or '' for x in self._get_raises_type()],
                key=lambda x: x or ''
            ) or None
        elif section == Sections.VARIABLES_SECTION:
            return self._sorted_keys(self._get_variable_type_lookup()) or None
        else:
            raise Exception(
                'Section type {} does not have items, '.format(
                    section.name
                ) + 'or is not yet supported.'
            )
        return None

    def get_noqas(self):
        # type: () -> Dict[str, List[str]]
        """Get a map of the errors ignored to their targets.

        Returns:
            A dictionary containing the errors to ignore as keys and
            a list of which targets to apply these exceptions to as
            the values.  A blank list implies a global noqa.

        """
        noqas = dict()
        for noqa in self._lookup[NoqaIdentifier.key]:
            noqas[NoqaIdentifier.extract(noqa) or '*'] = (
                NoqaIdentifier.extract_targets(noqa)
            )
        return noqas

    def get_line_numbers(self, node_type):
        # type: (str) -> Optional[Tuple[int, int]]
        """Get the line numbers for the first instance of the given section.

        Args:
            node_type: The NodeType which we want line numbers for.
                These should be unique instances. (I.e. they should be
                in the set of compound NodeTypes which only occur
                once in a docstring. For example, "Raises" and "Args".

        Returns:
            The line numbers for the first instance of the given node type.

        """
        nodes = self._lookup[node_type]
        if nodes:
            return nodes[0].line_numbers
        return None

    def get_line_numbers_for_value(self, node_type, value):
        # type: (str, str) -> Optional[Tuple[int, int]]
        """Get the line number for a node with the given value.

        Args:
            node_type: The compound node which should contain the
                node we are searching for.
            value: The value of the node.

        Returns:
            A list of line numbers for nodes which match the
            parameters.

        """
        nodes = self._lookup[node_type]
        for node in nodes:
            for child in node.walk():
                if child.value == value and child.line_numbers:
                    return child.line_numbers
        return None

    @property
    def ignore_all(self):
        # type: () -> bool
        """Return whether we should ignore everything in the docstring.

        This happens when there is a bare noqa in the docstring, or
        there is "noqa: *" in the docstring.

        Returns:
            True if we should ignore everything, otherwise false.

        """
        return False

    def get_style_errors(self):
        # type: () -> Iterable[Tuple[Callable, Tuple[int, int]]]
        """Get any style errors annotated on the tree.

        Yields:
            Instances of DarglintErrors for style issues.

        # noqa: I302

        """
        for node in self.root.in_order_traverse():
            for annotation in node.annotations:
                if issubclass(annotation, DarglintError):
                    yield annotation, node.line_numbers



import abc

from ..custom_assert import Assert
from .cyk import CykNode
from typing import (
    Callable,
    List,
    Optional,
    Union,
    Tuple,
)
from ..token import (
    TokenType,
)


class Continuation(object):
    """Represents a continuation of a path.

    This is the actual representation of the path in the parse
    tree.  The wrapper, Path, returns continuations.  Each
    continuation either performs a branching or a chain.

    """

    def __init__(self, path, condition, child=None):
        # type: (str, Callable[[CykNode], bool], Union[Continuation, Tuple[Continuation, ...], None]) -> None  # noqa: E501
        self.path = path
        self.condition = condition
        self.child = child
        self._sealed = False

    def of(self, path):
        # type: (str) -> Continuation
        Assert(
            not self._sealed,
            'Sealed continuations shouldn\'t be extended!',
        )
        if isinstance(self.child, Continuation):
            self.child.of(path)
        elif self.child is None:
            self.child = Continuation(path, lambda _: True, None)
        return self

    def branch(self, *continuations):
        # type: (Continuation) -> Continuation
        Assert(
            not self._sealed,
            'Sealed continuations shouldn\'t be extended!',
        )
        self.child = continuations
        self._sealed = True
        return self

    def extract(self, node):
        # type: (CykNode) -> Union[str, CykNode, None]
        """Extract the value of the leaf node described by this path.

        Args:
            node: The root of the path we're about to follow.

        Returns:
            The value of the token, if correctly described by this
            path.

        """
        if not self.condition(node):
            return None
        curr = node  # type: Optional[CykNode]
        for letter in self.path:
            if curr and letter == 'r':
                curr = curr.rchild
            elif curr and letter == 'l':
                curr = curr.lchild
            elif curr and letter == 'v' and curr.value:
                return curr.value.value
            else:
                return None
        if curr is None:
            return None

        is_branch = isinstance(self.child, tuple)
        is_continuation = isinstance(self.child, Continuation)
        if is_branch:
            for branch in self.child:  # type: ignore
                next_curr = branch.extract(curr)
                if isinstance(next_curr, str):
                    return next_curr
                elif next_curr is None:
                    # In branching, we try each branch until
                    # one succeeds.
                    continue
                elif isinstance(next_curr, CykNode):
                    return next_curr
                else:
                    Assert(
                        False,
                        'Expected path extraction to yield str '
                        'or None or CykNode but was {}'.format(
                            next_curr.__class__.__name__
                        )
                    )
                    return None
            # Branches are always terminal.  If we haven't
            # found the result by this point, there is none.
            return None
        elif is_continuation:
            next_curr = self.child.extract(curr)  # type: ignore
            if isinstance(next_curr, str):
                return next_curr
            elif next_curr is None:
                # In an unconditional chain, we fail if any
                # in the chain fail.
                return None
            else:
                Assert(
                    False,
                    'Expected path extraction to yield str '
                    'or None but was {}'.format(
                        next_curr.__class__.__name__
                    )
                )
                return None
        return curr


class Path(object):
    """Represents a path which can be taken in a parse tree.

    The purpose of this path is to extract the value of a
    token in the leaves of the tree.  We previously handled
    this my simply accessing the members of the CykNodes
    in the tree, and making asserts.  However, that can sometimes
    result in runtime errors.  Rather, we should want to
    log the failure in a type-safe manner.

    """

    @staticmethod
    def of(path):
        # type: (str) -> Continuation
        """Construct a path composed of left and right turns.

        If this is the terminal path, it should end with the
        character, 'v'.

        Args:
            path: The path to take.  Should be a string composed
                of the characters, 'l', 'r', and 'v'.

        Returns:
            A continuation of the path.

        """
        return Continuation(path, lambda _: True)

    @staticmethod
    def branch(*paths):
        # type: (Continuation) -> Continuation
        """A path which accepts the first succeeding path.

        Args:
            paths: The paths to try, in order.

        Returns:
            A continuation representing the path.

        """
        return Continuation('', lambda _: True, paths)

    # These methods are technically unnecessary -- they are
    # synonymous with an `of`.  However, it makes for nicer
    # documentation of intent, I think.
    @staticmethod
    def if_left(path):
        # type: (str) -> Continuation
        return Continuation(path, lambda x: bool(x.lchild))

    @staticmethod
    def if_right(path):
        # type: (str) -> Continuation
        return Continuation(path, lambda x: bool(x.rchild))


class Identifier(abc.ABC):
    """Identifies a particular target in the grammar.

    Since the transformation from BNF to CNF messes up
    the grammars, we can use this class to mark pieces
    of the grammar we need access to.  This should greatly
    simplify the process of interpreting the produced tree.

    """

    @staticmethod
    @property
    @abc.abstractmethod
    def key():
        # type: () -> str
        pass

    @staticmethod
    @property
    @abc.abstractmethod
    def path():
        # type: () -> Continuation
        pass

    @classmethod
    def extract(cls, node):
        # type: (CykNode) -> str
        # TODO: Fix the type annotation here.
        value = cls.path.extract(node)  # type: ignore
        Assert(
            value is not None,
            'Failed to extract {}'.format(cls.key)
        )
        return value or ''

    def __str__(self):
        return '<identifier: {}>'.format(self.key)

    def __repr__(self):
        return str(self)


class ArgumentItemIdentifier(Identifier):

    key = 'id_ArgsItem'
    path = Path.of('lv')


class ArgumentIdentifier(Identifier):

    key = 'id_Args'
    path = Path.of('rlv')


class ArgumentTypeIdentifier(Identifier):

    key = 'id_ArgType'
    path = Path.of('rrl').branch(
        Path.if_right('r').branch(Path.of('lv'), Path.of('l')),
        Path.of('v'),
        Path.of(''),
    )

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        value = ArgumentTypeIdentifier.path.extract(node)
        is_leaf = isinstance(value, str)
        is_branch = isinstance(value, CykNode)
        Assert(is_leaf or is_branch, 'Unable to extract argument type.')
        if is_leaf:
            return value  # type: ignore
        elif is_branch:
            return value.reconstruct_string()  # type: ignore
        else:
            return ''


class ExceptionItemIdentifier(Identifier):

    key = 'id_ExceptItem'
    path = Path.of('lv')


class ExceptionIdentifier(Identifier):

    key = 'id_Except'
    path = Path.of('rlv')


class ReturnTypeIdentifier(Identifier):

    key = 'id_ReturnType'
    path = Path.of('lv')


class YieldTypeIdentifier(Identifier):

    key = 'id_YieldType'
    path = Path.of('lv')


class NoqaIdentifier(Identifier):

    key = 'id_Noqa'
    path = Path.of('rr').branch(Path.of('v'), Path.if_left('lv'))

    @staticmethod
    def extract(node):
        # type: (CykNode) -> str
        if node.rchild and node.rchild.rchild:
            path = Path.branch(Path.of('rrv'), Path.of('rrlv'))
            value = path.extract(node)
            if isinstance(value, str):
                return value
            # path2 = Path.of('rrlv')
            # value = path2.extract(node)
            # if isinstance(value, str):
            #     return value
            return ''
        else:
            return ''

    @staticmethod
    def extract_targets(node):
        # type: (CykNode) -> List[str]
        if (
            node.rchild
            and node.rchild.rchild
        ):
            children = list()
            for child in node.rchild.rchild.walk():
                if child.value and child.value.token_type == TokenType.WORD:
                    children.append(child.value.value)
            if len(children) > 1:
                return children[1:]
        return []



import ast
from collections import (
    deque,
)
import copy
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    Union,
)
from ..config import get_logger
from ..custom_assert import Assert


logger = get_logger()



class Context(object):
    """A context which tracks exceptions and symbols."""

    def __init__(self):
        # type: () -> None
        self.exceptions = set()  # type: Set[str]

        # If we're in a bare handler, we have to capture new
        # exceptions raised separately from the existing ones.
        # So, we copy the existing exceptions over here.
        # This complicates the logic, for the calling class (as
        # contextual operations have to account for two cases),
        # but it doesn't seem avoidable.
        self.bare_handler_exceptions = None # type: Optional[Set[str]]

        # A lookup from variable names to AST nodes.
        # If the variable name occurs in a raise expression,
        # then the exception will be added using this lookup.
        self.variables = dict()  # type: Dict[str, Union[str, List[str]]]

        # The error(s) which the current exception block is
        # handling. (Since we only handle one handler at a time
        # in the context, and since they don't repeat the
        # exception, it's fine to overwrite this value.)
        self.handling = None  # type: Optional[List[str]]

    def set_in_bare_handler(self):
        self.bare_handler_exceptions = set(self.exceptions)
        self.remove_all_exceptions()

    def _get_attr_name(self, attr):
        # type: (Union[ast.Attribute, ast.Name, ast.Tuple]) -> List[str]
        curr = attr  # type: Any
        parts = list()  # type: List[str]

        # We assume here that the ast has a limited
        # depth.  Even if it's several thousand long,
        # it should work fine.
        while curr:
            if isinstance(curr, ast.Attribute):
                parts.append(curr.attr)
                curr = curr.value
            elif isinstance(curr, ast.Name):
                parts.append(curr.id)
                curr = None
            elif isinstance(curr, ast.Tuple):
                names = list()
                for node in curr.elts:
                    if isinstance(node, ast.Name):
                        names.extend(self._get_attr_name(node))
                    elif isinstance(node, ast.Attribute):
                        names.extend(self._get_attr_name(node))
                    else:
                        logger.error(
                            'While getting the names from a caught '
                            'tuple of exceptions, encountered '
                            'something other than an ast.Name: '
                            '{}'.format(
                                node.__class__.__name__
                            )
                        )
                return names
            else:
                logger.error(
                    'While getting ast.Attribute representation '
                    'a node had an unexpected type {}'.format(
                        curr.__class__.__name__
                    )
                )
                curr = None
        parts.reverse()
        return ['.'.join(parts)]

    def _get_name_name(self, name):
        # type: (Union[ast.Name, ast.Tuple]) -> Union[str, List[str]]
        if isinstance(name, ast.Name):
            return name.id
        elif isinstance(name, ast.Tuple):
            ret = list()
            for node in name.elts:
                if isinstance(node, ast.Name):
                    ret.append(node.id)
            return ret

    def _get_exception_name(self, raises):
        # type: (ast.Raise) -> Union[str, List[str]]
        if isinstance(raises, str):
            return raises
        if isinstance(raises.exc, ast.Name):
            # TODO: The name could come from a higher context --
            # we should check all parent contexts.
            name = raises.exc.id
            # The name was stored in the except statement,
            # so we should restore the type.
            if name in self.variables:
                return self.variables[name]
            else:
                return name
        elif isinstance(raises.exc, ast.Call):
            if hasattr(raises.exc.func, 'id'):
                return getattr(raises.exc.func, 'id')
            elif hasattr(raises.exc.func, 'attr'):
                return getattr(raises.exc.func, 'attr')
            else:
                logger.debug(
                    'Raises function call has neither id nor attr.'
                    'has only: %s' % str(dir(raises.exc.func))
                )
        elif isinstance(raises.exc, ast.Attribute):
            return raises.exc.attr
        elif isinstance(raises.exc, ast.Subscript):
            id_repr = ''
            if hasattr(raises.exc.value, 'id'):
                id_repr = getattr(raises.exc.value, 'id')
            n_repr = ''
            if hasattr(raises.exc.slice, 'value'):
                value = getattr(raises.exc.slice, 'value')
                if hasattr(value, 'n'):
                    n_repr = getattr(value, 'n')
            return '{}[{}]'.format(
                id_repr,
                n_repr,
            )
        elif raises.exc is None:
            if not self.handling:
                return ''
            elif len(self.handling) == 1:
                return self.handling[0]
            else:
                return self.handling
        else:
            logger.debug('Unexpected type in raises expression: {}'.format(
                raises.exc
            ))
        return ''

    def add_exception(self, node):
        # type: (ast.Raise) -> Set[str]
        """Add an exception to the context.

        If the exception(s) doesn't have a name and doesn't have
        more children, then it's a bare raise.  In that case, we
        return the exception(s) to the parent context.

        Args:
            node: A raise ast node.

        Returns:
            A list of exceptions to be passed up to the parent
            context.

        """
        name = self._get_exception_name(node)
        if name == '':
            if self.bare_handler_exceptions is not None:
                return self.bare_handler_exceptions | self.exceptions
            elif self.exceptions:
                return self.exceptions
            elif self.variables:
                values = set()
                for value in self.variables.values():
                    if isinstance(value, list):
                        values |= set(value)
                    else:
                        values.add(value)
                return values
            else:
                logger.warning(
                    'Unexpectedly had no exception name raised and no exception '
                    'in context.'
                )
        if isinstance(name, str):
            self.exceptions.add(name)
        elif isinstance(name, list):
            for part in name:
                self.exceptions.add(part)
        else:
            logger.warning('Node {} name extraction failed.')
        return set()

    def remove_exception(self, node):
        # type: (ast.Raise) -> None
        name = self._get_exception_name(node)
        if isinstance(name, str) and name in self.exceptions:
            self.exceptions.remove(name)
            self.handling = [name]
        elif isinstance(name, list):
            self.handling = []
            for part in name:
                self.exceptions.remove(part)
                self.handling.append(part)

    def remove_all_exceptions(self):
        # type: () -> None
        self.exceptions = set()

    def add_variable(self, variable, exception):
        # type: (str, Union[ast.Name, ast.Tuple]) -> None
        self.variables[variable] = self._get_name_name(exception)

    def set_handling(self, attr):
        # type: (Union[ast.Attribute, ast.Name, ast.Tuple]) -> None
        self.handling = self._get_attr_name(attr)

    def remove_variable(self, variable):
        # type: (str) -> None
        del self.variables[variable]

    def extend(self, other):
        # type: (Context) -> None
        self.exceptions |= other.exceptions

    def finish_handling(self):
        # type: () -> None
        self.handling = None


class RaiseVisitor(ast.NodeVisitor):

    def __init__(self, *args, **kwargs):
        # type: (Any, Any) -> None

        # Allow the raise visitor to be used in a mixin.
        # TODO: https://github.com/python/mypy/issues/4001
        super(RaiseVisitor, self).__init__(*args, **kwargs)  # type: ignore

        # The context in which an exception can be raised.
        # The default context is the function body,
        # and a new context is created for each try-except
        # statement.  When the current context is finished,
        # its errors are merged upwards.
        self.contexts = deque([Context()])

    @property
    def exceptions(self):
        # type: () -> Set[str]
        return self.contexts[0].exceptions

    @property
    def context(self):
        # type: () -> Context
        return self.contexts[-1]

    def visit_Raise(self, node):
        # type: (ast.Raise) -> ast.AST
        bubbles = self.context.add_exception(node)
        if bubbles:
            Assert(
                len(self.contexts) > 1,
                'We should only encounter bare raises in a handler.'
            )
            if len(self.contexts) < 2:
                return self.generic_visit(node)
            parent_context = self.contexts[-2]
            parent_context.exceptions |= bubbles

        return self.generic_visit(node)

    def visit_Try(self, node):
        # type: (ast.Try) -> None
        self.contexts.append(Context())
        for child in node.body:
            self.visit(child)
        for handler in node.handlers:
            if handler.type:
                if handler.name and (
                    isinstance(handler.type, ast.Name) or
                    isinstance(handler.type, ast.Tuple)
                ):
                    self.context.add_variable(handler.name, handler.type)
                elif isinstance(handler.type, ast.Attribute):
                    self.context.set_handling(handler.type)
                elif isinstance(handler.type, ast.Name):
                    self.context.set_handling(handler.type)
                elif isinstance(handler.type, ast.Tuple):
                    self.context.set_handling(handler.type)
                else:
                    logger.error(
                        'While getting the types of exceptions in '
                        'the handler, expected to find an ast.Name, '
                        'ast.Tuple, or ast.Attribute, but got {}'.format(
                            handler.type
                        )
                    )
                id = getattr(handler.type, 'id', None)
                if id:
                    self.context.remove_exception(id)

                self.generic_visit(handler)
                self.context.finish_handling()
            else:
                # Handle a bare except.
                #
                # Since the bare except handles all exceptions,
                # we have to clear all exceptions from the context.
                # However, exceptions could also be raised from
                # this handler.  So we can't clear the exceptions
                # first.  But if we clear the exceptions second,
                # then remove any new exceptions raised in the handler.
                # What we need, then, is to know which new exceptions
                # are raised, and clear all but them.  For that,
                # we use a temporary context.
                self.context.set_in_bare_handler()
                self.generic_visit(handler)

        for child in node.finalbody:
            self.visit(child)

        for child in node.orelse:
            self.visit(child)

        context = self.contexts.pop()
        self.context.extend(context)
